/* This file is part of the Views Template Library, written by
 * Gary Powell & Martin Weiser
 *
 * Copyright (c) 2000  Gary Powell
 * Copyright (c) 2000  Konrad Zuse Zentrum fuer Informationstechnik Berlin
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted 
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */

// This file contains class templates for ownership management
// and a suitable generator to select apropriate classes.

#if !defined(_VIEW_OWNERSHIP_DETAIL_HPP)
#define _VIEW_OWNERSHIP_DETAIL_HPP


#include <boost/type_traits.hpp>
#include <boost/smart_ptr.hpp>

BEGIN_VIEW_NAMESPACE

// A "smart pointer" that contains the "pointed to" value by
// aggregation. The template parameter X defines the value type that is
// visible from outside.

template <class X>
class aggregated {
  // Even if the visible type is a const one, we need a
  // non const type for aggregation in order to be
  // assignable and swappable.
  typedef typename boost::remove_const<X>::type value_type;
  value_type x;

public:
  explicit aggregated(): x() {}
  
  explicit aggregated(X const* x_): x(*x_) {}
  
  X* operator->() { return &x; }
  X const* operator->() const { return &x; }

  X& operator*() { return x; }
  X const& operator*() const { return x; }

  void swap(aggregated& a) { swap(x,a.x); }
};

END_VIEW_NAMESPACE

BEGIN_SWAP_NAMESPACE

template <class X>
void swap(VIEW_NAMESPACE::aggregated<X>& a,
          VIEW_NAMESPACE::aggregated<X>& b) {
  a.swap(b);
}

END_SWAP_NAMESPACE



BEGIN_VIEW_NAMESPACE

// The available tag classes denoting ownership semantics.
class external_ownership {};
class shared_ownership {};
class aggregated_ownership {};
class referenced_ownership {};


// A generator class for computing the suitable "smart pointer" type,
// depending on the constness tag and the ownership tag.

template <class X, class ownership_tag>
struct ownership_helper {};

template <class X> struct ownership_helper<X,external_ownership> {
  typedef X* type;
};
template <class X> struct ownership_helper<X,shared_ownership> {
  typedef boost::shared_ptr<X> type;
};
template <class X> struct ownership_helper<X,aggregated_ownership> {
  typedef aggregated<X> type;
};
template <class X> struct ownership_helper<X,referenced_ownership> {
  typedef boost::scoped_ptr<X> type;
};

                                          

template <class X, class const_tag, class ownership_tag>
struct ownership
  : public ownership_helper<X,ownership_tag> {};

template <class X, class ownership_tag>
struct ownership<X,const_view_tag,ownership_tag>
  : public ownership_helper<X const,ownership_tag> {};

//
// This is a traits class for determining the argument type a view's
// constructor should use for its container argument.
//
//                      | const_view_tag | mutable_view_tag
//----------------------+----------------+------------------
// aggregated_ownership |    const&      |    const&             <- we make a copy of our own
// referenced_ownership |    const&      |    const&             <- we make a copy of our own
// external_ownership   |    const&      |    &
// shared_ownership     |    const&      |    &
//
// Note that the container_type defined by view_traits does
// not satisfy this requirements. It does the exclusive ownership 
// cases wrong.
//

template <class container, 
          class const_tag, 
          class ownership_tag>
struct ctor_arg { typedef container const type; };

template <class container>
struct ctor_arg<container,mutable_view_tag,external_ownership> { typedef container type; };
template <class container>
struct ctor_arg<container,mutable_view_tag,shared_ownership> { typedef container type; };

END_VIEW_NAMESPACE


#endif
