/* This file is part of the Views Template Library, written by
 * Gary Powell & Martin Weiser
 *
 * Copyright (c) 2000  Gary Powell
 * Copyright (c) 2000  Konrad Zuse Zentrum fuer Informationstechnik Berlin
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */

#include "transform_view.hpp"
#include <numeric>
#include <vector>
#include <cstdlib>
#include <functional>
#include <cassert>
#include <ctime>
#include <iostream>
#include <cmath>

USING_VIEW_NAMESPACE

#include "performance.hpp"

typedef vector<pair<int,int> > Container;

//
// The following two functions are the ones which are timed.
// Given a container of integer pairs, they compute the largest
// first component occurring in the container.
// 
// The views approach uses a transform_view to project the pairs
// onto their first component and then uses max_element to find
// the largest one.
//
// The standard approach creates a parallel vector of int into
// which the first components are copied via transform. Then,
// max_element is used to find the largest one.
//

int views_run(Container const& c) {
  using namespace std;

  typedef transform_view<Container,select1st<pair<int,int> > > view;
  view data(c,select1st<pair<int,int> >());
  return *max_element(data.begin(),data.end());
}



int std_run(Container const& c) {
  using namespace std;

  vector<int> data(c.size());
  transform(c.begin(),c.end(),data.begin(),select1st<pair<int,int> >());
  return *max_element(data.begin(),data.end());
}

//
// The main program just triggers the measurement for container sizes
// from one up until the measurements take too long. This means the
// computation is probably stopped once swapping starts.
//



Container generator(unsigned long n) {
  Container x(n);
  for (unsigned long i=0; i<n; ++i)
    x[i] = make_pair(rand(),rand());
  return x;
}


int main(int argc, char* argv[]) {
  using namespace std;
  typedef pair<string,pointer_to_unary_function<Container const&,int> > Test;
  typedef vector<Test>    Tests;
  
  Tests tests;
  tests.push_back(Test("vtl",ptr_fun(views_run)));
  tests.push_back(Test("std",ptr_fun(std_run)));

  do_test_to_gnuplot(ptr_fun(generator),tests.begin(),tests.end(),1,600,cout);
  
  return 0;
}

// CVS $Id: performance_test.cpp,v 1.1 2000/06/13 16:11:32 bzfweise Exp $
